# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"get_nb_name": "000_export.ipynb",
         "get_colab_nb_name": "000_export.ipynb",
         "get_nb_path": "000_export.ipynb",
         "nb_name_to_py": "000_export.ipynb",
         "get_script_path": "000_export.ipynb",
         "nb2py": "000_export.ipynb",
         "is_nparray": "001_utils.ipynb",
         "is_tensor": "001_utils.ipynb",
         "is_zarr": "001_utils.ipynb",
         "is_dask": "001_utils.ipynb",
         "is_memmap": "001_utils.ipynb",
         "is_slice": "001_utils.ipynb",
         "totensor": "001_utils.ipynb",
         "toarray": "001_utils.ipynb",
         "toL": "001_utils.ipynb",
         "to3dtensor": "001_utils.ipynb",
         "to2dtensor": "001_utils.ipynb",
         "to1dtensor": "001_utils.ipynb",
         "to3darray": "001_utils.ipynb",
         "to2darray": "001_utils.ipynb",
         "to1darray": "001_utils.ipynb",
         "to3d": "001_utils.ipynb",
         "to2d": "001_utils.ipynb",
         "to1d": "001_utils.ipynb",
         "to2dPlus": "001_utils.ipynb",
         "to3dPlus": "001_utils.ipynb",
         "to2dPlusTensor": "001_utils.ipynb",
         "to2dPlusArray": "001_utils.ipynb",
         "to3dPlusTensor": "001_utils.ipynb",
         "to3dPlusArray": "001_utils.ipynb",
         "todtype": "001_utils.ipynb",
         "bytes2size": "001_utils.ipynb",
         "bytes2GB": "001_utils.ipynb",
         "get_size": "001_utils.ipynb",
         "get_dir_size": "001_utils.ipynb",
         "get_file_size": "001_utils.ipynb",
         "is_file": "001_utils.ipynb",
         "is_dir": "001_utils.ipynb",
         "delete_all_in_dir": "001_utils.ipynb",
         "reverse_dict": "001_utils.ipynb",
         "is_tuple": "001_utils.ipynb",
         "itemify": "001_utils.ipynb",
         "isnone": "001_utils.ipynb",
         "exists": "001_utils.ipynb",
         "ifelse": "001_utils.ipynb",
         "is_not_close": "001_utils.ipynb",
         "test_not_close": "001_utils.ipynb",
         "test_type": "001_utils.ipynb",
         "test_ok": "001_utils.ipynb",
         "test_not_ok": "001_utils.ipynb",
         "test_error": "001_utils.ipynb",
         "test_eq_nan": "001_utils.ipynb",
         "assert_fn": "001_utils.ipynb",
         "test_gt": "001_utils.ipynb",
         "test_ge": "001_utils.ipynb",
         "test_lt": "001_utils.ipynb",
         "test_le": "001_utils.ipynb",
         "stack": "001_utils.ipynb",
         "stack_pad": "001_utils.ipynb",
         "pad_sequences": "001_utils.ipynb",
         "match_seq_len": "001_utils.ipynb",
         "random_shuffle": "001_utils.ipynb",
         "cat2int": "001_utils.ipynb",
         "cycle_dl": "001_utils.ipynb",
         "cycle_dl_to_device": "001_utils.ipynb",
         "cycle_dl_estimate": "001_utils.ipynb",
         "cache_data": "001_utils.ipynb",
         "memmap2cache": "001_utils.ipynb",
         "cache_memmap": "001_utils.ipynb",
         "get_func_defaults": "001_utils.ipynb",
         "get_idx_from_df_col_vals": "001_utils.ipynb",
         "get_sublist_idxs": "001_utils.ipynb",
         "flatten_list": "001_utils.ipynb",
         "display_pd_df": "001_utils.ipynb",
         "ttest": "001_utils.ipynb",
         "kstest": "001_utils.ipynb",
         "tscore": "001_utils.ipynb",
         "pcc": "001_utils.ipynb",
         "scc": "001_utils.ipynb",
         "a": "001_utils.ipynb",
         "b": "001_utils.ipynb",
         "remove_fn": "001_utils.ipynb",
         "npsave": "001_utils.ipynb",
         "np_save": "001_utils.ipynb",
         "permute_2D": "001_utils.ipynb",
         "random_normal": "001_utils.ipynb",
         "random_half_normal": "001_utils.ipynb",
         "random_normal_tensor": "001_utils.ipynb",
         "random_half_normal_tensor": "001_utils.ipynb",
         "default_dpi": "001_utils.ipynb",
         "get_plot_fig": "001_utils.ipynb",
         "fig2buf": "001_utils.ipynb",
         "plot_scatter": "001_utils.ipynb",
         "get_idxs": "001_utils.ipynb",
         "apply_cmap": "001_utils.ipynb",
         "torch_tile": "001_utils.ipynb",
         "to_tsfresh_df": "001_utils.ipynb",
         "pcorr": "001_utils.ipynb",
         "scorr": "001_utils.ipynb",
         "torch_diff": "001_utils.ipynb",
         "get_outliers_IQR": "001_utils.ipynb",
         "clip_outliers": "001_utils.ipynb",
         "get_percentile": "001_utils.ipynb",
         "torch_clamp": "001_utils.ipynb",
         "get_robustscale_params": "001_utils.ipynb",
         "torch_slice_by_dim": "001_utils.ipynb",
         "torch_nanmean": "001_utils.ipynb",
         "torch_nanstd": "001_utils.ipynb",
         "concat": "001_utils.ipynb",
         "reduce_memory_usage": "001_utils.ipynb",
         "cls_name": "001_utils.ipynb",
         "roll2d": "001_utils.ipynb",
         "roll3d": "001_utils.ipynb",
         "random_roll2d": "001_utils.ipynb",
         "random_roll3d": "001_utils.ipynb",
         "rotate_axis0": "001_utils.ipynb",
         "rotate_axis1": "001_utils.ipynb",
         "rotate_axis2": "001_utils.ipynb",
         "chunks_calculator": "001_utils.ipynb",
         "is_memory_shared": "001_utils.ipynb",
         "assign_in_chunks": "001_utils.ipynb",
         "create_array": "001_utils.ipynb",
         "create_empty_array": "001_utils.ipynb",
         "np_save_compressed": "001_utils.ipynb",
         "np_load_compressed": "001_utils.ipynb",
         "np2memmap": "001_utils.ipynb",
         "torch_mean_groupby": "001_utils.ipynb",
         "torch_flip": "001_utils.ipynb",
         "torch_nan_to_num": "001_utils.ipynb",
         "torch_masked_to_num": "001_utils.ipynb",
         "mpl_trend": "001_utils.ipynb",
         "int2digits": "001_utils.ipynb",
         "array2digits": "001_utils.ipynb",
         "sincos_encoding": "001_utils.ipynb",
         "linear_encoding": "001_utils.ipynb",
         "encode_positions": "001_utils.ipynb",
         "sort_generator": "001_utils.ipynb",
         "get_subset_dict": "001_utils.ipynb",
         "create_dir": "001_utils.ipynb",
         "remove_dir": "001_utils.ipynb",
         "named_partial": "001_utils.ipynb",
         "yaml2dict": "001_utils.ipynb",
         "str2list": "001_utils.ipynb",
         "str2index": "001_utils.ipynb",
         "get_cont_cols": "001_utils.ipynb",
         "get_cat_cols": "001_utils.ipynb",
         "alphabet": "001_utils.ipynb",
         "ALPHABET": "001_utils.ipynb",
         "get_mapping": "001_utils.ipynb",
         "map_array": "001_utils.ipynb",
         "log_tfm": "001_utils.ipynb",
         "to_sincos_time": "001_utils.ipynb",
         "plot_feature_dist": "001_utils.ipynb",
         "rolling_moving_average": "001_utils.ipynb",
         "ffill_sequence": "001_utils.ipynb",
         "bfill_sequence": "001_utils.ipynb",
         "fbfill_sequence": "001_utils.ipynb",
         "dummify": "001_utils.ipynb",
         "shuffle_along_axis": "001_utils.ipynb",
         "analyze_feature": "001_utils.ipynb",
         "analyze_array": "001_utils.ipynb",
         "get_relpath": "001_utils.ipynb",
         "split_in_chunks": "001_utils.ipynb",
         "save_object": "001_utils.ipynb",
         "load_object": "001_utils.ipynb",
         "get_idxs_to_keep": "001_utils.ipynb",
         "zerofy": "001_utils.ipynb",
         "RandomSplitter": "010_data.validation.ipynb",
         "check_overlap": "010_data.validation.ipynb",
         "check_splits_overlap": "010_data.validation.ipynb",
         "leakage_finder": "010_data.validation.ipynb",
         "balance_idx": "010_data.validation.ipynb",
         "TrainValidTestSplitter": "010_data.validation.ipynb",
         "plot_splits": "010_data.validation.ipynb",
         "get_splits": "010_data.validation.ipynb",
         "TSSplitter": "010_data.validation.ipynb",
         "TimeSplitter": "010_data.validation.ipynb",
         "get_predefined_splits": "010_data.validation.ipynb",
         "combine_split_data": "010_data.validation.ipynb",
         "get_splits_len": "010_data.validation.ipynb",
         "prepare_idxs": "011_data.preparation.ipynb",
         "prepare_sel_vars_and_steps": "011_data.preparation.ipynb",
         "apply_sliding_window": "011_data.preparation.ipynb",
         "df2Xy": "011_data.preparation.ipynb",
         "split_Xy": "011_data.preparation.ipynb",
         "df2xy": "011_data.preparation.ipynb",
         "split_xy": "011_data.preparation.ipynb",
         "df2np3d": "011_data.preparation.ipynb",
         "add_missing_value_cols": "011_data.preparation.ipynb",
         "add_missing_timestamps": "011_data.preparation.ipynb",
         "time_encoding": "011_data.preparation.ipynb",
         "forward_gaps": "011_data.preparation.ipynb",
         "backward_gaps": "011_data.preparation.ipynb",
         "nearest_gaps": "011_data.preparation.ipynb",
         "get_gaps": "011_data.preparation.ipynb",
         "add_delta_timestamp_cols": "011_data.preparation.ipynb",
         "SlidingWindow": "011_data.preparation.ipynb",
         "SlidingWindowSplitter": "011_data.preparation.ipynb",
         "SlidingWindowPanel": "011_data.preparation.ipynb",
         "SlidingWindowPanelSplitter": "011_data.preparation.ipynb",
         "identify_padding": "011_data.preparation.ipynb",
         "decompress_from_url": "012_data.external.ipynb",
         "download_data": "012_data.external.ipynb",
         "get_UCR_univariate_list": "012_data.external.ipynb",
         "UTSC_datasets": "012_data.external.ipynb",
         "UCR_univariate_list": "012_data.external.ipynb",
         "get_UCR_multivariate_list": "012_data.external.ipynb",
         "MTSC_datasets": "012_data.external.ipynb",
         "UCR_multivariate_list": "012_data.external.ipynb",
         "UCR_list": "012_data.external.ipynb",
         "classification_list": "012_data.external.ipynb",
         "TSC_datasets": "012_data.external.ipynb",
         "get_UCR_data": "012_data.external.ipynb",
         "get_classification_data": "012_data.external.ipynb",
         "check_data": "012_data.external.ipynb",
         "get_Monash_regression_list": "012_data.external.ipynb",
         "Monash_regression_list": "012_data.external.ipynb",
         "regression_list": "012_data.external.ipynb",
         "TSR_datasets": "012_data.external.ipynb",
         "get_Monash_regression_data": "012_data.external.ipynb",
         "get_regression_data": "012_data.external.ipynb",
         "get_forecasting_list": "012_data.external.ipynb",
         "forecasting_time_series": "012_data.external.ipynb",
         "get_forecasting_time_series": "012_data.external.ipynb",
         "Monash_forecasting_list": "012_data.external.ipynb",
         "forecasting_list": "012_data.external.ipynb",
         "convert_tsf_to_dataframe": "012_data.external.ipynb",
         "get_Monash_forecasting_data": "012_data.external.ipynb",
         "get_forecasting_data": "012_data.external.ipynb",
         "NumpyTensor": "013_data.core.ipynb",
         "ToNumpyTensor": "013_data.core.ipynb",
         "TSTensor": "013_data.core.ipynb",
         "ToTSTensor": "013_data.core.ipynb",
         "show_tuple": "013_data.core.ipynb",
         "TSLabelTensor": "013_data.core.ipynb",
         "TSMaskTensor": "013_data.core.ipynb",
         "ToFloat": "013_data.core.ipynb",
         "ToInt": "013_data.core.ipynb",
         "TSClassification": "013_data.core.ipynb",
         "TSCategorize": "013_data.core.ipynb",
         "TSRegression": "013_data.core.ipynb",
         "TSForecasting": "013_data.core.ipynb",
         "TSMultiLabelClassification": "013_data.core.ipynb",
         "NumpyTensorBlock": "013_data.core.ipynb",
         "TSTensorBlock": "013_data.core.ipynb",
         "TorchDataset": "013_data.core.ipynb",
         "NumpyDataset": "013_data.core.ipynb",
         "TSDataset": "013_data.core.ipynb",
         "NoTfmLists": "013_data.core.ipynb",
         "TSTfmdLists": "013_data.core.ipynb",
         "NumpyDatasets": "013_data.core.ipynb",
         "tscoll_repr": "013_data.core.ipynb",
         "TSDatasets": "013_data.core.ipynb",
         "add_ds": "013_data.core.ipynb",
         "NumpyDatasets.add_dataset": "013_data.core.ipynb",
         "NumpyDatasets.add_test": "013_data.core.ipynb",
         "NumpyDatasets.add_unlabeled": "013_data.core.ipynb",
         "NumpyDataLoader": "013_data.core.ipynb",
         "TSDataLoader": "013_data.core.ipynb",
         "NumpyDataLoaders": "013_data.core.ipynb",
         "TSDataLoaders": "013_data.core.ipynb",
         "StratifiedSampler": "013_data.core.ipynb",
         "get_c": "013_data.core.ipynb",
         "get_best_dl_params": "013_data.core.ipynb",
         "get_best_dls_params": "013_data.core.ipynb",
         "get_ts_dls": "013_data.core.ipynb",
         "get_ts_dl": "013_data.core.ipynb",
         "get_subset_dl": "013_data.core.ipynb",
         "get_tsimage_dls": "013_data.core.ipynb",
         "get_time_per_batch": "013_data.core.ipynb",
         "get_dl_percent_per_epoch": "013_data.core.ipynb",
         "TSUnwindowedDataset": "014_data.unwindowed.ipynb",
         "TSUnwindowedDatasets": "014_data.unwindowed.ipynb",
         "TSMetaDataset": "015_data.metadatasets.ipynb",
         "TSMetaDatasets": "015_data.metadatasets.ipynb",
         "ToNumpyCategory": "016_data.preprocessing.ipynb",
         "OneHot": "016_data.preprocessing.ipynb",
         "TSNan2Value": "016_data.preprocessing.ipynb",
         "Nan2Value": "016_data.preprocessing.ipynb",
         "TSStandardize": "016_data.preprocessing.ipynb",
         "torch.Tensor.mul_min": "016_data.preprocessing.ipynb",
         "TSTensor.mul_min": "016_data.preprocessing.ipynb",
         "NumpyTensor.mul_min": "016_data.preprocessing.ipynb",
         "torch.Tensor.mul_max": "016_data.preprocessing.ipynb",
         "TSTensor.mul_max": "016_data.preprocessing.ipynb",
         "NumpyTensor.mul_max": "016_data.preprocessing.ipynb",
         "TSNormalize": "016_data.preprocessing.ipynb",
         "TSClipOutliers": "016_data.preprocessing.ipynb",
         "TSClip": "016_data.preprocessing.ipynb",
         "TSSelfMissingness": "016_data.preprocessing.ipynb",
         "TSRobustScale": "016_data.preprocessing.ipynb",
         "get_stats_with_uncertainty": "016_data.preprocessing.ipynb",
         "get_random_stats": "016_data.preprocessing.ipynb",
         "TSGaussianStandardize": "016_data.preprocessing.ipynb",
         "TSRandomStandardize": "016_data.preprocessing.ipynb",
         "TSDiff": "016_data.preprocessing.ipynb",
         "TSLog": "016_data.preprocessing.ipynb",
         "TSCyclicalPosition": "016_data.preprocessing.ipynb",
         "TSLinearPosition": "016_data.preprocessing.ipynb",
         "TSMissingness": "016_data.preprocessing.ipynb",
         "TSPositionGaps": "016_data.preprocessing.ipynb",
         "TSRollingMean": "016_data.preprocessing.ipynb",
         "TSLogReturn": "016_data.preprocessing.ipynb",
         "TSAdd": "016_data.preprocessing.ipynb",
         "TSClipByVar": "016_data.preprocessing.ipynb",
         "TSDropVars": "016_data.preprocessing.ipynb",
         "TSOneHotEncode": "016_data.preprocessing.ipynb",
         "TSPosition": "016_data.preprocessing.ipynb",
         "TSShrinkDataFrame": "016_data.preprocessing.ipynb",
         "TSOneHotEncoder": "016_data.preprocessing.ipynb",
         "TSCategoricalEncoder": "016_data.preprocessing.ipynb",
         "TSDateTimeEncoder": "016_data.preprocessing.ipynb",
         "default_date_attr": "016_data.preprocessing.ipynb",
         "TSMissingnessEncoder": "016_data.preprocessing.ipynb",
         "Preprocessor": "016_data.preprocessing.ipynb",
         "StandardScaler": "016_data.preprocessing.ipynb",
         "RobustScaler": "016_data.preprocessing.ipynb",
         "Normalizer": "016_data.preprocessing.ipynb",
         "BoxCox": "016_data.preprocessing.ipynb",
         "YeoJohnshon": "016_data.preprocessing.ipynb",
         "Quantile": "016_data.preprocessing.ipynb",
         "ReLabeler": "016_data.preprocessing.ipynb",
         "TSIdentity": "017_data.transforms.ipynb",
         "TSShuffle_HLs": "017_data.transforms.ipynb",
         "TSShuffleSteps": "017_data.transforms.ipynb",
         "TSGaussianNoise": "017_data.transforms.ipynb",
         "TSMagAddNoise": "017_data.transforms.ipynb",
         "TSMagMulNoise": "017_data.transforms.ipynb",
         "random_curve_generator": "017_data.transforms.ipynb",
         "random_cum_curve_generator": "017_data.transforms.ipynb",
         "random_cum_noise_generator": "017_data.transforms.ipynb",
         "random_cum_linear_generator": "017_data.transforms.ipynb",
         "TSTimeNoise": "017_data.transforms.ipynb",
         "TSMagWarp": "017_data.transforms.ipynb",
         "TSTimeWarp": "017_data.transforms.ipynb",
         "TSWindowWarp": "017_data.transforms.ipynb",
         "TSMagScale": "017_data.transforms.ipynb",
         "TSMagScalePerVar": "017_data.transforms.ipynb",
         "TSMagScaleByVar": "017_data.transforms.ipynb",
         "TSRandomResizedCrop": "017_data.transforms.ipynb",
         "TSRandomZoomIn": "017_data.transforms.ipynb",
         "TSWindowSlicing": "017_data.transforms.ipynb",
         "TSRandomZoomOut": "017_data.transforms.ipynb",
         "TSRandomTimeScale": "017_data.transforms.ipynb",
         "TSRandomTimeStep": "017_data.transforms.ipynb",
         "TSResampleSteps": "017_data.transforms.ipynb",
         "TSSubsampleSteps": "017_data.transforms.ipynb",
         "TSBlur": "017_data.transforms.ipynb",
         "TSSmooth": "017_data.transforms.ipynb",
         "maddest": "017_data.transforms.ipynb",
         "TSFreqDenoise": "017_data.transforms.ipynb",
         "TSRandomFreqNoise": "017_data.transforms.ipynb",
         "TSRandomResizedLookBack": "017_data.transforms.ipynb",
         "TSRandomLookBackOut": "017_data.transforms.ipynb",
         "TSVarOut": "017_data.transforms.ipynb",
         "TSCutOut": "017_data.transforms.ipynb",
         "TSTimeStepOut": "017_data.transforms.ipynb",
         "TSRandomCropPad": "017_data.transforms.ipynb",
         "TSMaskOut": "017_data.transforms.ipynb",
         "TSInputDropout": "017_data.transforms.ipynb",
         "TSTranslateX": "017_data.transforms.ipynb",
         "TSRandomShift": "017_data.transforms.ipynb",
         "TSHorizontalFlip": "017_data.transforms.ipynb",
         "TSRandomTrend": "017_data.transforms.ipynb",
         "TSRandomRotate": "017_data.transforms.ipynb",
         "TSVerticalFlip": "017_data.transforms.ipynb",
         "TSResize": "017_data.transforms.ipynb",
         "TSRandomSize": "017_data.transforms.ipynb",
         "TSRandomLowRes": "017_data.transforms.ipynb",
         "TSDownUpScale": "017_data.transforms.ipynb",
         "TSRandomDownUpScale": "017_data.transforms.ipynb",
         "TSRandomConv": "017_data.transforms.ipynb",
         "TSRandom2Value": "017_data.transforms.ipynb",
         "TSMask2Value": "017_data.transforms.ipynb",
         "all_TS_randaugs": "017_data.transforms.ipynb",
         "RandAugment": "017_data.transforms.ipynb",
         "TestTfm": "017_data.transforms.ipynb",
         "get_tfm_name": "017_data.transforms.ipynb",
         "MixHandler1d": "018_data.mixed_augmentation.ipynb",
         "MixUp1d": "018_data.mixed_augmentation.ipynb",
         "MixUp1D": "018_data.mixed_augmentation.ipynb",
         "CutMix1d": "018_data.mixed_augmentation.ipynb",
         "IntraClassCutMix1d": "018_data.mixed_augmentation.ipynb",
         "TSImage": "019_data.image.ipynb",
         "ToTSImage": "019_data.image.ipynb",
         "TSToPlot": "019_data.image.ipynb",
         "TSToMat": "019_data.image.ipynb",
         "TSToGADF": "019_data.image.ipynb",
         "TSToGASF": "019_data.image.ipynb",
         "TSToMTF": "019_data.image.ipynb",
         "TSToRP": "019_data.image.ipynb",
         "TSToJRP": "019_data.image.ipynb",
         "get_ts_features": "020_data.features.ipynb",
         "get_tabular_ds": "021_data.tabular.ipynb",
         "get_tabular_dls": "021_data.tabular.ipynb",
         "preprocess_df": "021_data.tabular.ipynb",
         "MixedDataLoader": "022_data.mixed.ipynb",
         "MixedDataLoaders": "022_data.mixed.ipynb",
         "get_mixed_dls": "022_data.mixed.ipynb",
         "HuberLoss": "050_losses.ipynb",
         "LogCoshLoss": "050_losses.ipynb",
         "MaskedLossWrapper": "050_losses.ipynb",
         "CenterLoss": "050_losses.ipynb",
         "CenterPlusLoss": "050_losses.ipynb",
         "FocalLoss": "050_losses.ipynb",
         "TweedieLoss": "050_losses.ipynb",
         "MatthewsCorrCoefBinary": "051_metrics.ipynb",
         "get_task_metrics": "051_metrics.ipynb",
         "accuracy_multi": "051_metrics.ipynb",
         "metrics_multi_common": "051_metrics.ipynb",
         "precision_multi": "051_metrics.ipynb",
         "recall_multi": "051_metrics.ipynb",
         "specificity_multi": "051_metrics.ipynb",
         "balanced_accuracy_multi": "051_metrics.ipynb",
         "Fbeta_multi": "051_metrics.ipynb",
         "F1_multi": "051_metrics.ipynb",
         "mae": "051_metrics.ipynb",
         "mape": "051_metrics.ipynb",
         "recall_at_specificity": "051_metrics.ipynb",
         "lift": "051_metrics.ipynb",
         "lift_at_specificity": "051_metrics.ipynb",
         "top_k_lift": "051_metrics.ipynb",
         "mean_per_class_accuracy": "051_metrics.ipynb",
         "Learner.show_batch": "052_learner.ipynb",
         "Learner.remove_all_cbs": "052_learner.ipynb",
         "Learner.one_batch": "052_learner.ipynb",
         "Learner.save_all": "052_learner.ipynb",
         "load_all": "052_learner.ipynb",
         "load_learner_all": "052_learner.ipynb",
         "Recorder.plot_metrics": "052_learner.ipynb",
         "Learner.plot_metrics": "052_learner.ipynb",
         "get_arch": "052_learner.ipynb",
         "all_archs_names": "052_learner.ipynb",
         "ts_learner": "052_learner.ipynb",
         "tsimage_learner": "052_learner.ipynb",
         "Learner.decoder": "052_learner.ipynb",
         "Learner.get_X_preds": "052a_inference.ipynb",
         "Learner.show_probas": "052b_analysis.ipynb",
         "Learner.plot_confusion_matrix": "052b_analysis.ipynb",
         "Learner.top_losses": "052b_analysis.ipynb",
         "Learner.plot_top_losses": "052b_analysis.ipynb",
         "Learner.feature_importance": "052b_analysis.ipynb",
         "Learner.step_importance": "052b_analysis.ipynb",
         "ModelWithTemperature": "052c_calibration.ipynb",
         "TemperatureSetter": "052c_calibration.ipynb",
         "ECELoss": "052c_calibration.ipynb",
         "plot_calibration_curve": "052c_calibration.ipynb",
         "Learner.calibrate_model": "052c_calibration.ipynb",
         "TSClassifier": "052d_tslearner.ipynb",
         "TSRegressor": "052d_tslearner.ipynb",
         "TSForecaster": "052d_tslearner.ipynb",
         "wrap_optimizer": "053_optimizer.ipynb",
         "TransformScheduler": "060_callback.core.ipynb",
         "ShowGraph": "060_callback.core.ipynb",
         "ShowGraphCallback2": "060_callback.core.ipynb",
         "SaveModel": "060_callback.core.ipynb",
         "get_lds_kernel_window": "060_callback.core.ipynb",
         "prepare_LDS_weights": "060_callback.core.ipynb",
         "WeightedPerSampleLoss": "060_callback.core.ipynb",
         "BatchSubsampler": "060_callback.core.ipynb",
         "GamblersCallback": "060_callback.experimental.ipynb",
         "gambler_loss": "060_callback.experimental.ipynb",
         "UBDAug": "060_callback.experimental.ipynb",
         "BatchLossFilter": "060_callback.experimental.ipynb",
         "RandomWeightLossWrapper": "060_callback.experimental.ipynb",
         "BatchMasker": "060_callback.experimental.ipynb",
         "SamplerWithReplacement": "060_callback.experimental.ipynb",
         "NoisyStudent": "061_callback.noisy_student.ipynb",
         "create_subsequence_mask": "063_callback.MVP.ipynb",
         "create_variable_mask": "063_callback.MVP.ipynb",
         "create_future_mask": "063_callback.MVP.ipynb",
         "self_mask": "063_callback.MVP.ipynb",
         "create_mask": "063_callback.MVP.ipynb",
         "MVP": "063_callback.MVP.ipynb",
         "TSBERT": "063_callback.MVP.ipynb",
         "PredictionDynamics": "064_callback.PredictionDynamics.ipynb",
         "init_lin_zero": "100_models.layers.ipynb",
         "lin_zero_init": "100_models.layers.ipynb",
         "SwishBeta": "100_models.layers.ipynb",
         "SmeLU": "100_models.layers.ipynb",
         "Chomp1d": "100_models.layers.ipynb",
         "same_padding1d": "100_models.layers.ipynb",
         "Pad1d": "100_models.layers.ipynb",
         "SameConv1d": "100_models.layers.ipynb",
         "same_padding2d": "100_models.layers.ipynb",
         "Pad2d": "100_models.layers.ipynb",
         "Conv2dSame": "100_models.layers.ipynb",
         "Conv2d": "100_models.layers.ipynb",
         "CausalConv1d": "100_models.layers.ipynb",
         "Conv1d": "100_models.layers.ipynb",
         "SeparableConv1d": "100_models.layers.ipynb",
         "AddCoords1d": "100_models.layers.ipynb",
         "ConvBlock": "100_models.layers.ipynb",
         "Conv": "100_models.layers.ipynb",
         "ConvBN": "100_models.layers.ipynb",
         "CoordConv": "100_models.layers.ipynb",
         "SepConv": "100_models.layers.ipynb",
         "ResBlock1dPlus": "100_models.layers.ipynb",
         "SEModule1d": "100_models.layers.ipynb",
         "Norm": "100_models.layers.ipynb",
         "BN1d": "100_models.layers.ipynb",
         "IN1d": "100_models.layers.ipynb",
         "LinLnDrop": "100_models.layers.ipynb",
         "LambdaPlus": "100_models.layers.ipynb",
         "Squeeze": "100_models.layers.ipynb",
         "Unsqueeze": "100_models.layers.ipynb",
         "Add": "100_models.layers.ipynb",
         "Concat": "100_models.layers.ipynb",
         "Permute": "100_models.layers.ipynb",
         "Transpose": "100_models.layers.ipynb",
         "View": "100_models.layers.ipynb",
         "Reshape": "100_models.layers.ipynb",
         "Max": "100_models.layers.ipynb",
         "LastStep": "100_models.layers.ipynb",
         "SoftMax": "100_models.layers.ipynb",
         "Clamp": "100_models.layers.ipynb",
         "Clip": "100_models.layers.ipynb",
         "ReZero": "100_models.layers.ipynb",
         "Noop": "100_models.layers.ipynb",
         "DropPath": "100_models.layers.ipynb",
         "Sharpen": "100_models.layers.ipynb",
         "Sequential": "122_models.TabFusionTransformer.ipynb",
         "TimeDistributed": "100_models.layers.ipynb",
         "Temp_Scale": "100_models.layers.ipynb",
         "Vector_Scale": "100_models.layers.ipynb",
         "Matrix_Scale": "100_models.layers.ipynb",
         "get_calibrator": "100_models.layers.ipynb",
         "LogitAdjustmentLayer": "100_models.layers.ipynb",
         "LogitAdjLayer": "100_models.layers.ipynb",
         "PPV": "100_models.layers.ipynb",
         "PPAuc": "100_models.layers.ipynb",
         "MaxPPVPool1d": "100_models.layers.ipynb",
         "AdaptiveWeightedAvgPool1d": "100_models.layers.ipynb",
         "GAP1d": "100_models.layers.ipynb",
         "GACP1d": "100_models.layers.ipynb",
         "GAWP1d": "100_models.layers.ipynb",
         "GlobalWeightedAveragePool1d": "100_models.layers.ipynb",
         "gwa_pool_head": "100_models.layers.ipynb",
         "GWAP1d": "100_models.layers.ipynb",
         "AttentionalPool1d": "100_models.layers.ipynb",
         "GAttP1d": "100_models.layers.ipynb",
         "attentional_pool_head": "100_models.layers.ipynb",
         "PoolingLayer": "100_models.layers.ipynb",
         "GEGLU": "123_models.TSPerceiver.ipynb",
         "ReGLU": "100_models.layers.ipynb",
         "get_act_fn": "100_models.layers.ipynb",
         "pytorch_acts": "100_models.layers.ipynb",
         "pytorch_act_names": "100_models.layers.ipynb",
         "create_pool_head": "100_models.layers.ipynb",
         "pool_head": "100_models.layers.ipynb",
         "average_pool_head": "100_models.layers.ipynb",
         "concat_pool_head": "100_models.layers.ipynb",
         "max_pool_head": "100_models.layers.ipynb",
         "create_pool_plus_head": "100_models.layers.ipynb",
         "pool_plus_head": "100_models.layers.ipynb",
         "create_conv_head": "100_models.layers.ipynb",
         "conv_head": "100_models.layers.ipynb",
         "create_mlp_head": "100_models.layers.ipynb",
         "mlp_head": "100_models.layers.ipynb",
         "create_fc_head": "100_models.layers.ipynb",
         "fc_head": "100_models.layers.ipynb",
         "create_rnn_head": "100_models.layers.ipynb",
         "rnn_head": "100_models.layers.ipynb",
         "imputation_head": "100_models.layers.ipynb",
         "create_conv_lin_nd_head": "100_models.layers.ipynb",
         "conv_lin_nd_head": "100_models.layers.ipynb",
         "conv_lin_3d_head": "100_models.layers.ipynb",
         "create_conv_lin_3d_head": "100_models.layers.ipynb",
         "create_lin_nd_head": "100_models.layers.ipynb",
         "lin_nd_head": "100_models.layers.ipynb",
         "lin_3d_head": "100_models.layers.ipynb",
         "create_lin_3d_head": "100_models.layers.ipynb",
         "create_conv_3d_head": "100_models.layers.ipynb",
         "conv_3d_head": "100_models.layers.ipynb",
         "universal_pool_head": "100_models.layers.ipynb",
         "heads": "100_models.layers.ipynb",
         "SqueezeExciteBlock": "100_models.layers.ipynb",
         "GaussianNoise": "100_models.layers.ipynb",
         "PositionwiseFeedForward": "100_models.layers.ipynb",
         "TokenLayer": "100_models.layers.ipynb",
         "ScaledDotProductAttention": "123_models.TSPerceiver.ipynb",
         "MultiheadAttention": "100_models.layers.ipynb",
         "MultiConv1d": "100_models.layers.ipynb",
         "LSTMOutput": "100_models.layers.ipynb",
         "emb_sz_rule": "100_models.layers.ipynb",
         "TSEmbedding": "100_models.layers.ipynb",
         "MultiEmbedding": "100_models.layers.ipynb",
         "apply_idxs": "100b_models.utils.ipynb",
         "SeqTokenizer": "100b_models.utils.ipynb",
         "SeqEmbed": "100b_models.utils.ipynb",
         "get_embed_size": "100b_models.utils.ipynb",
         "get_layers": "100b_models.utils.ipynb",
         "is_layer": "100b_models.utils.ipynb",
         "is_linear": "100b_models.utils.ipynb",
         "is_bn": "100b_models.utils.ipynb",
         "is_conv_linear": "100b_models.utils.ipynb",
         "is_affine_layer": "100b_models.utils.ipynb",
         "is_conv": "100b_models.utils.ipynb",
         "has_bias": "100b_models.utils.ipynb",
         "has_weight": "100b_models.utils.ipynb",
         "has_weight_or_bias": "100b_models.utils.ipynb",
         "check_bias": "100b_models.utils.ipynb",
         "check_weight": "100b_models.utils.ipynb",
         "get_nf": "100b_models.utils.ipynb",
         "ts_splitter": "100b_models.utils.ipynb",
         "transfer_weights": "100b_models.utils.ipynb",
         "build_ts_model": "100b_models.utils.ipynb",
         "build_model": "100b_models.utils.ipynb",
         "create_model": "100b_models.utils.ipynb",
         "count_parameters": "100b_models.utils.ipynb",
         "build_tsimage_model": "100b_models.utils.ipynb",
         "build_tabular_model": "100b_models.utils.ipynb",
         "create_tabular_model": "100b_models.utils.ipynb",
         "get_clones": "100b_models.utils.ipynb",
         "split_model": "100b_models.utils.ipynb",
         "output_size_calculator": "100b_models.utils.ipynb",
         "change_model_head": "100b_models.utils.ipynb",
         "naive_forecaster": "100b_models.utils.ipynb",
         "true_forecaster": "100b_models.utils.ipynb",
         "PositionalEncoding": "100c_models.positional_encoders.ipynb",
         "SinCosPosEncoding": "100c_models.positional_encoders.ipynb",
         "Coord2dPosEncoding": "100c_models.positional_encoders.ipynb",
         "Coord1dPosEncoding": "100c_models.positional_encoders.ipynb",
         "get_acts_and_grads": "100d_models.explainability.ipynb",
         "get_attribution_map": "100d_models.explainability.ipynb",
         "ResBlock": "101_models.ResNet.ipynb",
         "ResNet": "101_models.ResNet.ipynb",
         "ResBlockPlus": "101b_models.ResNetPlus.ipynb",
         "ResNetPlus": "101b_models.ResNetPlus.ipynb",
         "InceptionModule": "102_models.InceptionTime.ipynb",
         "InceptionBlock": "102_models.InceptionTime.ipynb",
         "InceptionTime": "102_models.InceptionTime.ipynb",
         "InceptionModulePlus": "102b_models.InceptionTimePlus.ipynb",
         "InceptionBlockPlus": "102b_models.InceptionTimePlus.ipynb",
         "InceptionTimePlus": "102b_models.InceptionTimePlus.ipynb",
         "InCoordTime": "102b_models.InceptionTimePlus.ipynb",
         "XCoordTime": "102b_models.InceptionTimePlus.ipynb",
         "InceptionTimePlus17x17": "102b_models.InceptionTimePlus.ipynb",
         "InceptionTimePlus32x32": "102b_models.InceptionTimePlus.ipynb",
         "InceptionTimePlus47x47": "102b_models.InceptionTimePlus.ipynb",
         "InceptionTimePlus62x62": "102b_models.InceptionTimePlus.ipynb",
         "InceptionTimeXLPlus": "102b_models.InceptionTimePlus.ipynb",
         "MultiInceptionTimePlus": "102b_models.InceptionTimePlus.ipynb",
         "MLP": "103_models.MLP.ipynb",
         "FCN": "103b_models.FCN.ipynb",
         "FCNPlus": "103c_models.FCNPlus.ipynb",
         "gMLP": "103d_models.gMLP.ipynb",
         "ResCNN": "104_models.ResCNN.ipynb",
         "RNN": "105_models.RNN.ipynb",
         "LSTM": "105_models.RNN.ipynb",
         "GRU": "105_models.RNN.ipynb",
         "RNNPlus": "105_models.RNNPlus.ipynb",
         "LSTMPlus": "105_models.RNNPlus.ipynb",
         "GRUPlus": "105_models.RNNPlus.ipynb",
         "XceptionModule": "106_models.XceptionTime.ipynb",
         "XceptionBlock": "106_models.XceptionTime.ipynb",
         "XceptionTime": "106_models.XceptionTime.ipynb",
         "XceptionModulePlus": "106b_models.XceptionTimePlus.ipynb",
         "XceptionBlockPlus": "106b_models.XceptionTimePlus.ipynb",
         "XceptionTimePlus": "106b_models.XceptionTimePlus.ipynb",
         "RNN_FCN": "107_models.RNN_FCN.ipynb",
         "LSTM_FCN": "107_models.RNN_FCN.ipynb",
         "GRU_FCN": "107_models.RNN_FCN.ipynb",
         "MRNN_FCN": "107_models.RNN_FCN.ipynb",
         "MLSTM_FCN": "107_models.RNN_FCN.ipynb",
         "MGRU_FCN": "107_models.RNN_FCN.ipynb",
         "RNN_FCNPlus": "107b_models.RNN_FCNPlus.ipynb",
         "LSTM_FCNPlus": "107b_models.RNN_FCNPlus.ipynb",
         "GRU_FCNPlus": "107b_models.RNN_FCNPlus.ipynb",
         "MRNN_FCNPlus": "107b_models.RNN_FCNPlus.ipynb",
         "MLSTM_FCNPlus": "107b_models.RNN_FCNPlus.ipynb",
         "MGRU_FCNPlus": "107b_models.RNN_FCNPlus.ipynb",
         "TransformerModel": "108_models.TransformerModel.ipynb",
         "get_activation_fn": "108b_models.TST.ipynb",
         "TST": "108b_models.TST.ipynb",
         "TSTPlus": "108c_models.TSTPlus.ipynb",
         "MultiTSTPlus": "108c_models.TSTPlus.ipynb",
         "SampaddingConv1D_BN": "109_models.OmniScaleCNN.ipynb",
         "build_layer_with_layer_parameter": "109_models.OmniScaleCNN.ipynb",
         "OmniScaleCNN": "109_models.OmniScaleCNN.ipynb",
         "get_Prime_number_in_a_range": "109_models.OmniScaleCNN.ipynb",
         "get_out_channel_number": "109_models.OmniScaleCNN.ipynb",
         "generate_layer_parameter_list": "109_models.OmniScaleCNN.ipynb",
         "WaveBlock": "110_models.mWDN.ipynb",
         "mWDN": "110_models.mWDN.ipynb",
         "mWDNBlocks": "110_models.mWDN.ipynb",
         "mWDNPlus": "110_models.mWDN.ipynb",
         "RocketClassifier": "111_models.ROCKET.ipynb",
         "load_rocket": "111_models.ROCKET.ipynb",
         "RocketRegressor": "111_models.ROCKET.ipynb",
         "ROCKET": "111a_models.ROCKET_Pytorch.ipynb",
         "create_rocket_features": "111a_models.ROCKET_Pytorch.ipynb",
         "get_rocket_features": "111a_models.ROCKET_Pytorch.ipynb",
         "MiniRocketClassifier": "111b_models.MINIROCKET.ipynb",
         "load_minirocket": "111b_models.MINIROCKET.ipynb",
         "MiniRocketRegressor": "111b_models.MINIROCKET.ipynb",
         "MiniRocketVotingClassifier": "111b_models.MINIROCKET.ipynb",
         "get_minirocket_preds": "111b_models.MINIROCKET.ipynb",
         "MiniRocketVotingRegressor": "111b_models.MINIROCKET.ipynb",
         "MiniRocketFeatures": "111c_models.MINIROCKET_Pytorch.ipynb",
         "MRF": "111c_models.MINIROCKET_Pytorch.ipynb",
         "get_minirocket_features": "111d_models.MINIROCKETPlus_Pytorch.ipynb",
         "MiniRocketHead": "111d_models.MINIROCKETPlus_Pytorch.ipynb",
         "MiniRocket": "111c_models.MINIROCKET_Pytorch.ipynb",
         "MiniRocketFeaturesPlus": "111d_models.MINIROCKETPlus_Pytorch.ipynb",
         "Flatten": "111d_models.MINIROCKETPlus_Pytorch.ipynb",
         "MiniRocketPlus": "111d_models.MINIROCKETPlus_Pytorch.ipynb",
         "InceptionRocketFeaturesPlus": "111d_models.MINIROCKETPlus_Pytorch.ipynb",
         "InceptionRocketPlus": "111d_models.MINIROCKETPlus_Pytorch.ipynb",
         "xresnet1d18": "112_models.XResNet1d.ipynb",
         "xresnet1d34": "112_models.XResNet1d.ipynb",
         "xresnet1d50": "112_models.XResNet1d.ipynb",
         "xresnet1d101": "112_models.XResNet1d.ipynb",
         "xresnet1d152": "112_models.XResNet1d.ipynb",
         "xresnet1d18_deep": "112_models.XResNet1d.ipynb",
         "xresnet1d34_deep": "112_models.XResNet1d.ipynb",
         "xresnet1d50_deep": "112_models.XResNet1d.ipynb",
         "xresnet1d18_deeper": "112_models.XResNet1d.ipynb",
         "xresnet1d34_deeper": "112_models.XResNet1d.ipynb",
         "xresnet1d50_deeper": "112_models.XResNet1d.ipynb",
         "XResNet1dPlus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d18plus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d34plus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d50plus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d101plus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d152plus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d18_deepplus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d34_deepplus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d50_deepplus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d18_deeperplus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d34_deeperplus": "112b_models.XResNet1dPlus.ipynb",
         "xresnet1d50_deeperplus": "112b_models.XResNet1dPlus.ipynb",
         "TemporalBlock": "113_models.TCN.ipynb",
         "TemporalConvNet": "113_models.TCN.ipynb",
         "TCN": "113_models.TCN.ipynb",
         "XCM": "114_models.XCM.ipynb",
         "XCMPlus": "114b_models.XCMPlus.ipynb",
         "TabModel": "120_models.TabModel.ipynb",
         "TabBackbone": "120_models.TabModel.ipynb",
         "TabHead": "120_models.TabModel.ipynb",
         "ifnone": "122_models.TabFusionTransformer.ipynb",
         "SharedEmbedding": "121_models.TabTransformer.ipynb",
         "FullEmbeddingDropout": "121_models.TabTransformer.ipynb",
         "TabTransformer": "121_models.TabTransformer.ipynb",
         "GatedTabTransformer": "121b_models.GatedTabTransformer.ipynb",
         "TabFusionBackbone": "122_models.TabFusionTransformer.ipynb",
         "TabFusionTransformer": "122_models.TabFusionTransformer.ipynb",
         "TSTabFusionTransformer": "122_models.TabFusionTransformer.ipynb",
         "Attention": "123_models.TSPerceiver.ipynb",
         "FeedForward": "123_models.TSPerceiver.ipynb",
         "CrossAttention": "123_models.TSPerceiver.ipynb",
         "LatentTransformer": "123_models.TSPerceiver.ipynb",
         "TSPerceiver": "123_models.TSPerceiver.ipynb",
         "TSiTPlus": "124_models.TSiTPlus.ipynb",
         "TSiT": "124_models.TSiTPlus.ipynb",
         "TSSequencerPlus": "125_models.TSSequencerPlus.ipynb",
         "TSSequencer": "125_models.TSSequencerPlus.ipynb",
         "MultiInputNet": "130_models.MultiInputNet.ipynb",
         "InputWrapper": "140_models.misc.ipynb",
         "ResidualWrapper": "140_models.misc.ipynb",
         "RecursiveWrapper": "140_models.misc.ipynb",
         "run_optuna_study": "200_optuna.ipynb",
         "wandb_agent": "201_wandb.ipynb",
         "get_wandb_agent": "201_wandb.ipynb",
         "run_wandb_agent": "201_wandb.ipynb"}

modules = ["export.py",
           "utils.py",
           "data/validation.py",
           "data/preparation.py",
           "data/external.py",
           "data/core.py",
           "data/unwindowed.py",
           "data/metadatasets.py",
           "data/preprocessing.py",
           "data/transforms.py",
           "data/mixed_augmentation.py",
           "data/image.py",
           "data/features.py",
           "data/tabular.py",
           "data/mixed.py",
           "losses.py",
           "metrics.py",
           "learner.py",
           "inference.py",
           "analysis.py",
           "calibration.py",
           "tslearner.py",
           "optimizer.py",
           "callback/core.py",
           "callback/experimental.py",
           "callback/noisy_student.py",
           "callback/MVP.py",
           "callback/PredictionDynamics.py",
           "models/layers.py",
           "models/utils.py",
           "models/positional_encoders.py",
           "models/explainability.py",
           "models/ResNet.py",
           "models/ResNetPlus.py",
           "models/InceptionTime.py",
           "models/InceptionTimePlus.py",
           "models/MLP.py",
           "models/FCN.py",
           "models/FCNPlus.py",
           "models/gMLP.py",
           "models/ResCNN.py",
           "models/RNN.py",
           "models/RNNPlus.py",
           "models/XceptionTime.py",
           "models/XceptionTimePlus.py",
           "models/RNN_FCN.py",
           "models/RNN_FCNPlus.py",
           "models/TransformerModel.py",
           "models/TST.py",
           "models/TSTPlus.py",
           "models/OmniScaleCNN.py",
           "models/mWDN.py",
           "models/ROCKET.py",
           "models/ROCKET_Pytorch.py",
           "models/MINIROCKET.py",
           "models/MINIROCKET_Pytorch.py",
           "models/MINIROCKETPlus_Pytorch.py",
           "models/XResNet1d.py",
           "models/XResNet1dPlus.py",
           "models/TCN.py",
           "models/XCM.py",
           "models/XCMPlus.py",
           "models/TabModel.py",
           "models/TabTransformer.py",
           "models/GatedTabTransformer.py",
           "models/TabFusionTransformer.py",
           "models/TSPerceiver.py",
           "models/TSiTPlus.py",
           "models/TSSequencerPlus.py",
           "models/MultiInputNet.py",
           "models/misc.py",
           "optuna.py",
           "wandb.py"]

doc_url = "https://timeseriesAI.github.io/tsai/"

git_url = "https://github.com/timeseriesAI/tsai/tree/main/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/064_callback.PredictionDynamics.ipynb.

# %% auto 0
__all__ = ['PredictionDynamics']

# %% ../../nbs/064_callback.PredictionDynamics.ipynb 3
from fastai.callback.all import *
from ..imports import *

# %% ../../nbs/064_callback.PredictionDynamics.ipynb 4
class PredictionDynamics(Callback):
    order, run_valid = 65, True

    def __init__(self, show_perc=1., figsize=(10,6), alpha=.3, size=30, color='lime', cmap='gist_rainbow', normalize=False, 
                 sensitivity=None, specificity=None):

        """
        Args:
            show_perc:   percent of samples from the valid set that will be displayed. Default: 1 (all).
                         You can reduce it if the number is too high and the chart is too busy.
            alpha:       level of transparency. Default:.3. 1 means no transparency.
            figsize:     size of the chart. You may want to expand it if too many classes.
            size:        size of each sample in the chart. Default:30. You may need to decrease it a bit if too many classes/ samples.
            color:       color used in regression plots.
            cmap:        color map used in classification plots.
            normalize:   flag to normalize histograms displayed in binary classification.
            sensitivity: (aka recall or True Positive Rate) if you pass a float between 0. and 1. the sensitivity threshold will be plotted in the chart.
                         Only used in binary classification.
            specificity: (or True Negative Rate) if you pass a float between 0. and 1. it will be plotted in the chart. Only used in binary classification.

        The red line in classification tasks indicate the average probability of true class.
        """

        store_attr()

    def before_fit(self):
        self.run = not hasattr(self.learn, 'lr_finder') and not hasattr(self, "gather_preds")
        if not self.run:
            return
        self.cat = True if (hasattr(self.dls, "c") and self.dls.c > 1) else False
        if self.cat:
            self.binary = self.dls.c == 2
        if self.show_perc != 1:
            valid_size = len(self.dls.valid.dataset)
            self.show_idxs = np.random.choice(valid_size, int(round(self.show_perc * valid_size)), replace=False)

        # Prepare ground truth container
        self.y_true = []

    def before_epoch(self):
        # Prepare empty pred container in every epoch
        self.y_pred = []

    def after_pred(self):
        if self.training:
            return

        # Get y_true in epoch 0
        if self.epoch == 0:
            self.y_true.extend(self.y.cpu().flatten().numpy())

        # Gather y_pred for every batch
        if self.cat:
            if self.binary:
                y_pred = F.softmax(self.pred, -1)[:, 1].reshape(-1, 1).cpu()
            else:
                y_pred = torch.gather(F.softmax(self.pred, -1), -1, self.y.reshape(-1, 1).long()).cpu()
        else:
            y_pred = self.pred.cpu()
        self.y_pred.extend(y_pred.flatten().numpy())

    def after_epoch(self):
        # Ground truth
        if self.epoch == 0:
            self.y_true = np.array(self.y_true)
            if self.show_perc != 1:
                self.y_true = self.y_true[self.show_idxs]
            self.y_bounds = (np.min(self.y_true), np.max(self.y_true))
            self.min_x_bounds, self.max_x_bounds = np.min(self.y_true), np.max(self.y_true)

        self.y_pred = np.array(self.y_pred)
        if self.show_perc != 1:
            self.y_pred = self.y_pred[self.show_idxs]
        if self.cat:
            neg_thr = None
            pos_thr = None
            if self.specificity is not None: 
                inp0 = self.y_pred[self.y_true == 0]
                neg_thr = np.sort(inp0)[-int(len(inp0) * (1 - self.specificity))]
            if self.sensitivity is not None: 
                inp1 = self.y_pred[self.y_true == 1]
                pos_thr = np.sort(inp1)[-int(len(inp1) * self.sensitivity)]
            self.update_graph(self.y_pred, self.y_true, neg_thr=neg_thr, pos_thr=pos_thr)
        else:
            # Adjust bounds during validation
            self.min_x_bounds = min(self.min_x_bounds, np.min(self.y_pred))
            self.max_x_bounds = max(self.max_x_bounds, np.max(self.y_pred))
            x_bounds = (self.min_x_bounds, self.max_x_bounds)
            self.update_graph(self.y_pred, self.y_true, x_bounds=x_bounds, y_bounds=self.y_bounds)

    def update_graph(self, y_pred, y_true, x_bounds=None, y_bounds=None, neg_thr=None, pos_thr=None):
        if not hasattr(self, 'graph_fig'):
            self.df_out = display("", display_id=True)
            if self.cat:
                self._cl_names = self.dls.vocab
                self._classes = L(self.dls.vocab.o2i.values())
                self._n_classes = len(self._classes)
                if self.binary:
                    self.bins = np.linspace(0, 1, 101)
                else:
                    _cm = plt.get_cmap(self.cmap)
                    self._color = [_cm(1. * c/self._n_classes) for c in range(1, self._n_classes + 1)][::-1]
                    self._h_vals = np.linspace(-.5, self._n_classes - .5, self._n_classes + 1)[::-1]
                    self._rand = []
                    for i, c in enumerate(self._classes):
                        self._rand.append(.5 * (np.random.rand(np.sum(y_true == c)) - .5))
            self.graph_fig, self.graph_ax = plt.subplots(1, figsize=self.figsize)
            self.graph_out = display("", display_id=True)
        self.graph_ax.clear()
        if self.cat:
            if self.binary:
                self.graph_ax.hist(y_pred[y_true == 0], bins=self.bins, density=self.normalize,  color='red', label=self._cl_names[0],
                                          edgecolor='black', alpha=self.alpha)
                self.graph_ax.hist(y_pred[y_true == 1], bins=self.bins, density=self.normalize,  color='blue', label=self._cl_names[1],
                                          edgecolor='black', alpha=self.alpha)
                self.graph_ax.axvline(.5, lw=1, ls='--', color='gray')
                if neg_thr is not None: 
                    self.graph_ax.axvline(neg_thr, lw=2, ls='--', color='red', label=f'specificity={(self.specificity):.3f}')
                if pos_thr is not None: 
                    self.graph_ax.axvline(pos_thr, lw=2, ls='--', color='blue', label=f'sensitivity={self.sensitivity:.3f}')
                self.graph_ax.set_xlabel(f'probability of class {self._cl_names[1]}', fontsize=12)
                self.graph_ax.legend()
            else:
                for i, c in enumerate(self._classes):
                    self.graph_ax.scatter(y_pred[y_true == c], y_true[y_true == c] + self._rand[i],  color=self._color[i],
                                          edgecolor='black', alpha=self.alpha, lw=.5, s=self.size)
                    self.graph_ax.vlines(np.mean(y_pred[y_true == c]), i - .5, i + .5, color='r')
                self.graph_ax.vlines(.5, min(self._h_vals), max(self._h_vals), lw=.5)
                self.graph_ax.hlines(self._h_vals, 0, 1, lw=.5)
                self.graph_ax.set_ylim(min(self._h_vals), max(self._h_vals))
                self.graph_ax.set_yticks(self._classes)
                self.graph_ax.set_yticklabels(self._cl_names)
                self.graph_ax.set_ylabel('true class', fontsize=12)
                self.graph_ax.set_xlabel('probability of true class', fontsize=12)
            self.graph_ax.set_xlim(0, 1)
            self.graph_ax.set_xticks(np.linspace(0, 1, 11))
            self.graph_ax.grid(axis='x', color='gainsboro', lw=.2)
        else:
            self.graph_ax.scatter(y_pred, y_true, color=self.color, edgecolor='black', alpha=self.alpha, lw=.5, s=self.size)
            self.graph_ax.set_xlim(*x_bounds)
            self.graph_ax.set_ylim(*y_bounds)
            self.graph_ax.plot([*x_bounds], [*x_bounds], color='gainsboro')
            self.graph_ax.set_xlabel('y_pred', fontsize=12)
            self.graph_ax.set_ylabel('y_true', fontsize=12)
            self.graph_ax.grid(color='gainsboro', lw=.2)
        self.graph_ax.set_title(f'Prediction Dynamics \nepoch: {self.epoch + 1}/{self.n_epoch}')
        self.df_out.update(pd.DataFrame(np.stack(self.learn.recorder.values)[-1].reshape(1,-1),
                                        columns=self.learn.recorder.metric_names[1:-1], index=[self.epoch]))
        self.graph_out.update(self.graph_ax.figure)
        if self.epoch == self.n_epoch - 1: 
            plt.close(self.graph_ax.figure)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/103b_models.FCN.ipynb.

# %% auto 0
__all__ = ['FCN']

# %% ../../nbs/103b_models.FCN.ipynb 3
from ..imports import *
from .layers import *

# %% ../../nbs/103b_models.FCN.ipynb 4
class FCN(Module):
    def __init__(self, c_in, c_out, layers=[128, 256, 128], kss=[7, 5, 3]):
        assert len(layers) == len(kss)
        self.convblock1 = ConvBlock(c_in, layers[0], kss[0])
        self.convblock2 = ConvBlock(layers[0], layers[1], kss[1])
        self.convblock3 = ConvBlock(layers[1], layers[2], kss[2])
        self.gap = GAP1d(1)
        self.fc = nn.Linear(layers[-1], c_out)

    def forward(self, x):
        x = self.convblock1(x)
        x = self.convblock2(x)
        x = self.convblock3(x)
        x = self.gap(x)
        return self.fc(x)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/052b_analysis.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/052b_analysis.ipynb 3
from .imports import *
import sklearn.metrics as skm
from sklearn.model_selection import train_test_split
from fastai.learner import * 
from fastai.interpret import *
from .utils import *
from .data.preprocessing import *
from .data.core import *
from .inference import *

# %% ../nbs/052b_analysis.ipynb 4
@patch
@delegates(subplots)
def show_probas(self:Learner, figsize=(6,6), ds_idx=1, dl=None, one_batch=False, max_n=None, **kwargs):
    recorder = copy(self.recorder) # This is to avoid loss of recorded values while generating preds
    if dl is None: dl = self.dls[ds_idx]
    if one_batch: dl = [dl.one_batch()]
    probas, targets = self.get_preds(dl=dl)
    if probas.ndim == 2 and probas.min() < 0 or probas.max() > 1: probas = nn.Softmax(-1)(probas)
    if not isinstance(targets[0].item(), Integral): return
    targets = targets.flatten()
    if max_n is not None:
        idxs = np.random.choice(len(probas), max_n, False)
        probas, targets = probas[idxs], targets[idxs]
    if isinstance(probas, torch.Tensor): probas = probas.detach().cpu().numpy()
    if isinstance(targets, torch.Tensor): targets = targets.detach().cpu().numpy()
    fig = plt.figure(figsize=figsize, **kwargs)
    classes = np.unique(targets)
    nclasses = len(classes)
    vals = np.linspace(.5, .5 + nclasses - 1, nclasses)[::-1]
    plt.vlines(.5, min(vals) - 1, max(vals), color='black', linewidth=.5)
    cm = plt.get_cmap('gist_rainbow')
    color = [cm(1.* c/nclasses) for c in range(1, nclasses + 1)][::-1]
    # class_probas = np.array([probas[i,t] for i,t in enumerate(targets)])
    class_probas = np.array([probas[i][t] for i,t in enumerate(targets)])
    for i, c in enumerate(classes):
        plt.scatter(class_probas[targets == c] if nclasses > 2 or i > 0 else 1 - class_probas[targets == c],
                    targets[targets == c] + .5 * (np.random.rand((targets == c).sum()) - .5), color=color[i], edgecolor='black', alpha=.2, s=100)
        if nclasses > 2: plt.vlines((targets == c).mean(), i - .5, i + .5, color='r', linewidth=.5)
    plt.hlines(vals, 0, 1)
    plt.ylim(min(vals) - 1, max(vals))
    plt.xlim(0,1)
    plt.xticks(np.linspace(0,1,11), fontsize=12)
    plt.yticks(classes, [self.dls.vocab[x] for x in classes], fontsize=12)
    plt.title('Predicted proba per true class' if nclasses > 2 else 'Predicted class 1 proba per true class', fontsize=14)
    plt.xlabel('Probability', fontsize=12)
    plt.ylabel('True class', fontsize=12)
    plt.grid(axis='x', color='gainsboro', linewidth=.2)
    plt.show()
    self.recorder = recorder

# %% ../nbs/052b_analysis.ipynb 5
@patch
def plot_confusion_matrix(self:Learner, ds_idx=1, dl=None, thr=.5, normalize=False, title='Confusion matrix', cmap="Blues", norm_dec=2, figsize=(5,5),
                          title_fontsize=12, fontsize=10, plot_txt=True, **kwargs):
        "Plot the confusion matrix, with `title` and using `cmap`."
        # This function is mainly copied from the sklearn docs
        if dl is None: dl = self.dls[ds_idx]
        assert dl.cat
        if dl.c == 2: # binary classification
            probas, preds = self.get_preds(dl=dl)
            y_pred = (probas[:, 1] > thr).numpy().astype(int)
            y_test = preds.numpy()
            if normalize: skm_normalize = 'true'
            else: skm_normalize = None
            cm = skm.confusion_matrix(y_test, y_pred, normalize=skm_normalize)
        else: 
            cm = ClassificationInterpretation.from_learner(self).confusion_matrix()

        if normalize: cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        fig = plt.figure(figsize=figsize, **kwargs)
        plt.imshow(cm, interpolation='nearest', cmap=cmap)
        if self.dls.c == 2:
            plt.title(f"{title} (threshold: {thr})", fontsize=title_fontsize)
        else: 
            plt.title(title, fontsize=title_fontsize)
        tick_marks = np.arange(len(self.dls.vocab))
        plt.xticks(tick_marks, self.dls.vocab, rotation=90, fontsize=fontsize)
        plt.yticks(tick_marks, self.dls.vocab, rotation=0, fontsize=fontsize)

        if plot_txt:
            thresh = cm.max() / 2.
            for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
                coeff = f'{cm[i, j]:.{norm_dec}f}' if normalize else f'{cm[i, j]}'
                plt.text(j, i, coeff, horizontalalignment="center", verticalalignment="center", color="white" if cm[i, j] > thresh else "black", fontsize=fontsize)

        ax = fig.gca()
        ax.set_ylim(len(self.dls.vocab)-.5,-.5)

        plt.tight_layout()
        plt.ylabel('Actual', fontsize=fontsize)
        plt.xlabel('Predicted', fontsize=fontsize)
        plt.grid(False)

# %% ../nbs/052b_analysis.ipynb 6
@patch
def top_losses(self:Learner,
    X, # array-like object representing the independent variables
    y, # array-like object representing the target
    k:int=9, # Optional. #items to plot
    largest=True, # Flag to show largest or smallest losses
    bs:int=64, # batch size
    ):
    *_, losses = self.get_X_preds(X, y, bs=bs, with_loss=True)
    top_losses, idxs = losses.topk(ifnone(k, len(losses)), largest=largest)
    idxs = idxs.tolist()
    return top_losses, idxs

@patch
def plot_top_losses(self:Learner,
    X, # array-like object representing the independent variables
    y, # array-like object representing the target
    k:int=9, # Optional. #items to plot
    largest=True, # Flag to show largest or smallest losses
    bs:int=64, # batch size
    **kwargs, # show_batch kwargs
    ):
    *_, losses = self.get_X_preds(X, y, bs=bs, with_loss=True)
    idxs = losses.topk(ifnone(k, len(losses)), largest=largest)[1].tolist()
    dl = self.dls.valid.new_dl(X[idxs], y=y[idxs], bs=k)
    b = dl.one_batch()
    dl.show_batch(b, max_n=k, **kwargs)

# %% ../nbs/052b_analysis.ipynb 9
@patch
def feature_importance(self:Learner, 
    X=None, # array-like object containing the time series. If None, all data in the validation set will be used.
    y=None, # array-like object containing the targets. If None, all targets in the validation set will be used.
    partial_n:(int, float)=None, # # (int) or % (float) of used to measure feature importance. If None, all data will be used.
    method:str='permutation', # Method used to invalidate feature. Use 'permutation' for shuffling or 'ablation' for setting values to np.nan.
    feature_names:list=None, # Optional list of feature names that will be displayed if available. Otherwise var_0, var_1, etc.
    sel_classes:(str, list)=None, # classes for which the analysis will be made
    key_metric_idx:int=0, # Optional position of the metric used. If None or no metric is available, the loss will be used.
    show_chart:bool=True, # Flag to indicate if a chart showing permutation feature importance will be plotted.
    figsize:tuple=None, # Size of the chart.
    title:str=None, # Optional string that will be used as the chart title. If None 'Permutation Feature Importance'.
    return_df:bool=True, # Flag to indicate if the dataframe with feature importance will be returned.
    save_df_path:Path=None, # Path where dataframe containing the permutation feature importance results will be saved.
    random_state:int=23, # Optional int that controls the shuffling applied to the data.
    verbose:bool=True, # Flag that controls verbosity.
    ):
    r"""Calculates feature importance as the drop in the model's validation loss or metric when a feature value is randomly shuffled"""
    
    global output, metric
    
    assert method in ['permutation', 'ablation']

    # X, y
    if X is None:
        X = self.dls.train.dataset.tls[0].items
        if hasattr(self.dls.train.dataset.tls[0], '_splits'): X = X[self.dls.train.dataset.tls[0]._splits]
    if y is None:
        y = self.dls.train.dataset.tls[1].items
    if partial_n is not None:
        _, rand_idxs, *_ = train_test_split(np.arange(len(y)), y, test_size=partial_n, random_state=random_state, stratify=y)
        X = X.oindex[rand_idxs] if hasattr(X, 'oindex') else X[rand_idxs]
        y = y.oindex[rand_idxs] if hasattr(y, 'oindex') else y[rand_idxs]
    else: 
        X, y = X[:], y[:]
    if sel_classes is not None:
        filt = np.isin(y, listify(sel_classes))
        X, y = X[filt], y[filt]
    pv(f'X.shape: {X.shape}', verbose)
    pv(f'y.shape: {y.shape}', verbose)

    # Metrics
    metrics = [mn for mn in self.recorder.metric_names if mn not in ['epoch', 'train_loss', 'valid_loss', 'time']]
    if len(metrics) == 0 or key_metric_idx is None:
        metric_name = self.loss_func.__class__.__name__
        key_metric_idx = None
    else:
        metric_name = metrics[key_metric_idx]
        metric = self.recorder.metrics[key_metric_idx].func
    metric_name = metric_name.replace("train_", "").replace("valid_", "")
    pv(f'Selected metric: {metric_name}', verbose)

    # Selected vars & feature names
    sel_vars = not(isinstance(self.dls.sel_vars, slice) and self.dls.sel_vars == slice(None, None, None))
    if feature_names is None:
        feature_names = L([f"var_{i}" for i in range(X.shape[1])])
        if sel_vars:
            feature_names = feature_names[self.dls.sel_vars]
    else:
        feature_names = listify(feature_names)

    if sel_vars:
        assert len(feature_names) == len(self.dls.sel_vars)
    else:
        assert len(feature_names) == X.shape[1]
    sel_var_idxs = L(np.arange(X.shape[1]).tolist())
    if sel_vars:
        sel_var_idxs = sel_var_idxs[self.dls.sel_vars]
    assert len(feature_names) == len(sel_var_idxs)
    g = list(zip(np.arange(len(sel_var_idxs)+2), [0] + sel_var_idxs))
    
    # Loop
    COLS = ['BASELINE'] + list(feature_names)
    results = []
    pv(f'Computing feature importance ({method} method)...', verbose)
    try:
        if method == 'ablation':
            fs = self.dls.valid.after_batch.fs
            self.dls.valid.after_batch.fs = fs + [TSNan2Value()]
        for i,k in progress_bar(g):
            if i > 0:
                if k not in sel_var_idxs: continue
                save_feat = X[:, k].copy()
                if method == 'permutation':
                    # shuffle along samples & steps
                    X[:, k] = random_shuffle(X[:, k].flatten(), random_state=random_state).reshape(X[:, k].shape)
                elif method == 'ablation':
                    X[:, k] = np.nan
            if key_metric_idx is None:
                value = self.get_X_preds(X, y, with_loss=True)[-1].mean().item()
            else:
                output = self.get_X_preds(X, y)
                if self.dls.c == 2:
                    try: value = metric(output[0][:, 1], output[1]).item()
                    except: value = metric(output[0], output[1]).item()
                else:
                    value = metric(output[0], output[1]).item()
                del output
            pv(f"{k:3} feature: {COLS[i]:20} {metric_name}: {value:8.6f}", verbose)
            results.append([COLS[i], value])
            del value; gc.collect()
            if i > 0:
                X[:, k] = save_feat
                del save_feat; gc.collect()
        
        if method == 'ablation':
            self.dls.valid.after_batch.fs = fs

    except KeyboardInterrupt:
        if i > 0:
            X[:, k] = save_feat
            del save_feat; gc.collect()
        if method == 'ablation':
            self.dls.valid.after_batch.fs = fs

    # DataFrame
    df = pd.DataFrame(results, columns=["Feature", metric_name])
    df[f'{metric_name}_change'] = df[metric_name] - df.loc[0, metric_name]
    sign = np.sign(df[f'{metric_name}_change'].mean())
    if sign == 0: sign = 1
    df[f'{metric_name}_change'] = df[f'{metric_name}_change'] * sign

    # Display feature importance
    if show_chart:
        print()
        value_change = df.loc[1:, f'{metric_name}_change'].values
        pos_value_change = value_change.copy()
        neg_value_change = value_change.copy()
        pos_value_change[pos_value_change < 0] = 0
        neg_value_change[neg_value_change > 0] = 0
        if figsize is None:
            figsize=(10, .5*len(value_change))
        plt.figure(figsize=figsize)
        plt.barh(np.arange(len(value_change))[::-1], pos_value_change, color='lime', edgecolor='black')
        plt.barh(np.arange(len(value_change))[::-1], neg_value_change, color='red', edgecolor='black')
        plt.axvline(0, color='black')
        plt.yticks(np.arange(len(value_change))[::-1], df.loc[1:, "Feature"].values)
        if title is None: title = f'Feature Importance ({method} method)'
        plt.title(title, size=16)
        text = 'increase' if sign == 1 else 'decrease'
        plt.xlabel(f"{metric_name} {text} when feature is removed")
        plt.ylim((-1,len(value_change)))
        plt.show()

    # Save feature importance
    df = df.sort_values(metric_name, ascending=sign < 0).reset_index(drop=True)
    if save_df_path:
        if save_df_path.split('.')[-1] != 'csv': save_df_path = f'{save_df_path}.csv'
        df.to_csv(f'{save_df_path}', index=False)
        pv(f'Feature importance df saved to {save_df_path}', verbose)
    if return_df: 
        return df 

# %% ../nbs/052b_analysis.ipynb 10
@patch
def step_importance(
    self:Learner, 
    X=None, # array-like object containing the time series. If None, all data in the validation set will be used.
    y=None, # array-like object containing the targets. If None, all targets in the validation set will be used.
    partial_n:(int, float)=None, # # (int) or % (float) of used to measure feature importance. If None, all data will be used.
    method:str='permutation', # Method used to invalidate feature. Use 'permutation' for shuffling or 'ablation' for setting values to np.nan.
    step_names:list=None, # Optional list of step names that will be displayed if available. Otherwise 0, 1, 2, etc.
    sel_classes:(str, list)=None, # classes for which the analysis will be made
    n_steps:int=1, # # of steps that will be analyzed at a time. Default is 1.
    key_metric_idx:int=0, # Optional position of the metric used. If None or no metric is available, the loss will be used.
    show_chart:bool=True, # Flag to indicate if a chart showing permutation feature importance will be plotted.
    figsize:tuple=(10, 5), # Size of the chart.
    title:str=None, # Optional string that will be used as the chart title. If None 'Permutation Feature Importance'.
    xlabel=None, # Optional string that will be used as the chart xlabel. If None 'steps'.
    return_df:bool=True, # Flag to indicate if the dataframe with feature importance will be returned.
    save_df_path:Path=None, # Path where dataframe containing the permutation feature importance results will be saved.
    random_state:int=23, # Optional int that controls the shuffling applied to the data.
    verbose:bool=True, # Flag that controls verbosity.
    ):
    r"""Calculates step importance as the drop in the model's validation loss or metric when a step/s value/s is/are randomly shuffled"""

    assert method in ['permutation', 'ablation']
    
    # X, y
    if X is None:
        X = self.dls.train.dataset.tls[0].items
        if hasattr(self.dls.train.dataset.tls[0], '_splits'): X = X[self.dls.train.dataset.tls[0]._splits]
    if y is None:
        y = self.dls.train.dataset.tls[1].items
    if partial_n is not None:
        _, rand_idxs, *_ = train_test_split(np.arange(len(y)), y, test_size=partial_n, random_state=random_state, stratify=y)
        X = X.oindex[rand_idxs] if hasattr(X, 'oindex') else X[rand_idxs]
        y = y.oindex[rand_idxs] if hasattr(y, 'oindex') else y[rand_idxs]
    else: 
        X, y = X[:], y[:]
    if sel_classes is not None:
        filt = np.isin(y, listify(sel_classes))
        X, y = X[filt], y[filt]
    pv(f'X.shape: {X.shape}', verbose)
    pv(f'y.shape: {y.shape}', verbose)

    # Metrics
    metrics = [mn for mn in self.recorder.metric_names if mn not in ['epoch', 'train_loss', 'valid_loss', 'time']]
    if len(metrics) == 0 or key_metric_idx is None:
        metric_name = self.loss_func.__class__.__name__
        key_metric_idx = None
    else:
        metric_name = metrics[key_metric_idx]
        metric = self.recorder.metrics[key_metric_idx].func
    metric_name = metric_name.replace("train_", "").replace("valid_", "")
    pv(f'Selected metric: {metric_name}', verbose)
    
    # Selected steps
    sel_step_idxs = L(np.arange(X.shape[-1]).tolist())[self.dls.sel_steps]
    if n_steps != 1:
        sel_step_idxs = [listify(sel_step_idxs[::-1][n:n+n_steps][::-1]) for n in range(0, len(sel_step_idxs), n_steps)][::-1]     
    g = list(zip(np.arange(len(sel_step_idxs)+2), [0] + sel_step_idxs))

    # Loop
    COLS = ['BASELINE'] + sel_step_idxs
    results = []
    _step_names = []
    pv('Computing step importance...', verbose)
    try:
        if method == 'ablation':
            fs = self.dls.valid.after_batch.fs
            self.dls.valid.after_batch.fs = fs + [TSNan2Value()]
        for i,k in progress_bar(g):
            if i > 0:
                if k not in sel_step_idxs: continue
                save_feat = X[..., k].copy()
                if method == 'permutation':
                    # shuffle along samples
                    X[..., k] = shuffle_along_axis(X[..., k], axis=0, random_state=random_state)
                elif method == 'ablation':
                    X[..., k] = np.nan
            if key_metric_idx is None:
                value = self.get_X_preds(X, y, with_loss=True)[-1].mean().item()
            else:
                output = self.get_X_preds(X, y)
                if self.dls.c == 2:
                    try: value = metric(output[0][:, 1], output[1]).item()
                    except: value = metric(output[0], output[1]).item()
                else:
                    value = metric(output[0], output[1]).item()
                del output
            
            # Step names
            if i == 0 or step_names is None:
                if i > 0 and n_steps != 1:
                    step_name = f"{str(COLS[i][0])} to {str(COLS[i][-1])}"
                else: step_name = str(COLS[i])
            else:
                step_name = step_names[i - 1]
            if i > 0: _step_names.append(step_name)
                
            pv(f"{i:3} step: {step_name:20} {metric_name}: {value:8.6f}", verbose)
            results.append([step_name, value])
            del value; gc.collect()
            if i > 0:
                X[..., k] = save_feat
                del save_feat; gc.collect()
        
        if method == 'ablation':
            self.dls.valid.after_batch.fs = fs

    except KeyboardInterrupt:
        if i > 0:
            X[..., k] = save_feat
            del save_feat; gc.collect()
        if method == 'ablation':
            self.dls.valid.after_batch.fs = fs

    # DataFrame
    df = pd.DataFrame(results, columns=["Step", metric_name])
    df[f'{metric_name}_change'] = df[metric_name] - df.loc[0, metric_name]
    sign = np.sign(df[f'{metric_name}_change'].mean())
    if sign == 0: sign = 1
    df[f'{metric_name}_change'] = df[f'{metric_name}_change'] * sign
    
    # Display step importance
    if show_chart:
        print()
        value_change = df.loc[1:, f'{metric_name}_change'].values
        pos_value_change = value_change.copy()
        neg_value_change = value_change.copy()
        pos_value_change[pos_value_change < 0] = 0
        neg_value_change[neg_value_change > 0] = 0
        plt.figure(figsize=figsize)
        plt.bar(np.arange(len(value_change)), pos_value_change, color='lime', edgecolor='black')
        plt.bar(np.arange(len(value_change)), neg_value_change, color='red', edgecolor='black')
        plt.axhline(0, color='black')
        plt.xticks(np.arange(len(value_change)), _step_names, rotation=90)
        if title is None: title = f'Step Importance ({method} method)'
        plt.title(title, size=16)
        text = 'increase' if sign == 1 else 'decrease'
        if xlabel is None: xlabel = 'steps'
        plt.xlabel(xlabel)
        plt.ylabel(f"{metric_name} {text} when removed")
        plt.xlim((-1,len(value_change)))
        plt.show()

    # Save step importance
    df = df.sort_values(metric_name, ascending=sign < 0).reset_index(drop=True)
    if save_df_path:
        if save_df_path.split('.')[-1] != 'csv': save_df_path = f'{save_df_path}.csv'
        df.to_csv(f'{save_df_path}', index=False)
        pv(f'Step importance df saved to {save_df_path}', verbose)
    if return_df: 
        return df



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/010_data.validation.ipynb.

# %% auto 0
__all__ = ['TimeSplitter', 'RandomSplitter', 'check_overlap', 'check_splits_overlap', 'leakage_finder', 'balance_idx',
           'TrainValidTestSplitter', 'plot_splits', 'get_splits', 'get_walk_forward_splits', 'TSSplitter',
           'get_predefined_splits', 'combine_split_data', 'get_splits_len']

# %% ../../nbs/010_data.validation.ipynb 3
from ..imports import *
from imblearn.over_sampling import RandomOverSampler
from matplotlib.patches import Patch
from matplotlib.colors import LinearSegmentedColormap
from sklearn.model_selection import train_test_split, KFold, StratifiedKFold
from ..utils import *

# %% ../../nbs/010_data.validation.ipynb 4
def RandomSplitter(valid_pct=0.2, seed=None):
    "Create function that splits `items` between train/val with `valid_pct` randomly."
    def _inner(o):
        if seed is not None: torch.manual_seed(seed)
        rand_idx = L(list(torch.randperm(len(o)).numpy()))
        cut = int(valid_pct * len(o))
        return rand_idx[cut:],rand_idx[:cut]
    return _inner

# %% ../../nbs/010_data.validation.ipynb 5
def check_overlap(a, b, c=None):
    a = toarray(a)
    b = toarray(b)
    overlap_ab = np.isin(a, b)
    if c is None:
        if isinstance(overlap_ab[0], (list, L, np.ndarray, torch.Tensor)):  overlap_ab = overlap_ab[0] 
        if not any(overlap_ab): return False
        else: return a[overlap_ab].tolist()
    else:
        c = toarray(c)
        overlap_ac = np.isin(a, c)
        if isinstance(overlap_ac[0], (list, L, np.ndarray, torch.Tensor)):  overlap_ac = overlap_ac[0] 
        overlap_bc = np.isin(b, c)
        if isinstance(overlap_bc[0], (list, L, np.ndarray, torch.Tensor)):  overlap_bc = overlap_bc[0] 
        if not any(overlap_ab) and not any(overlap_ac) and not any(overlap_bc): return False
        else: return a[overlap_ab].tolist(), a[overlap_ac].tolist(), b[overlap_bc].tolist()

def check_splits_overlap(splits):
    return [check_overlap(*_splits) for _splits in splits] if is_listy(splits[0][0]) else check_overlap(*splits)

def leakage_finder(*splits, verbose=True):
    '''You can pass splits as a tuple, or train, valid, ...'''
    splits = L(*splits)
    overlaps = 0
    for i in range(len(splits)):
        for j in range(i + 1, len(splits)):
            overlap = check_overlap(splits[i], splits[j])
            if overlap: 
                pv(f'overlap between splits [{i}, {j}] {overlap}', verbose)
                overlaps += 1
    assert overlaps == 0, 'Please, review your splits!'

def balance_idx(o, shuffle=False, random_state=None, verbose=False):
    if isinstance(o, list): o = L(o)
    idx_ = np.arange(len(o)).reshape(-1, 1)
    ros = RandomOverSampler(random_state=random_state)
    resampled_idxs, _ = ros.fit_resample(idx_, np.asarray(o))
    new_idx = L(resampled_idxs.reshape(-1,).tolist())
    if shuffle: new_idx = random_shuffle(new_idx)
    return new_idx

# %% ../../nbs/010_data.validation.ipynb 12
def TrainValidTestSplitter(n_splits:int=1, valid_size:Union[float, int]=0.2, test_size:Union[float, int]=0., train_only:bool=False,
                           stratify:bool=True, balance:bool=False, shuffle:bool=True, random_state:Union[None, int]=None, verbose:bool=False, **kwargs):
    "Split `items` into random train, valid (and test optional) subsets."
    
    if not shuffle and stratify and not train_only: 
        pv('stratify set to False because shuffle=False. If you want to stratify set shuffle=True', verbose)
        stratify = False
        
    def _inner(o, **kwargs):
        if stratify:
            _, unique_counts = np.unique(o, return_counts=True)
            if np.min(unique_counts) >= 2 and np.min(unique_counts) >= n_splits: stratify_ = stratify  
            elif np.min(unique_counts) < n_splits: 
                stratify_ = False
                pv(f'stratify set to False as n_splits={n_splits} cannot be greater than the min number of members in each class ({np.min(unique_counts)}).', 
                   verbose)
            else:
                stratify_ = False
                pv('stratify set to False as the least populated class in o has only 1 member, which is too few.', verbose)
        else: stratify_ = False
        vs = 0 if train_only else 1. / n_splits if n_splits > 1 else int(valid_size * len(o)) if isinstance(valid_size, float) else valid_size
        if test_size: 
            ts = int(test_size * len(o)) if isinstance(test_size, float) else test_size
            train_valid, test = train_test_split(range(len(o)), test_size=ts, stratify=o if stratify_ else None, shuffle=shuffle, 
                                                 random_state=random_state, **kwargs)
            test = toL(test)
            if shuffle: test = random_shuffle(test, random_state)
            if vs == 0:
                train, _ = RandomSplitter(0, seed=random_state)(o[train_valid])
                train = toL(train)
                if balance: train = train[balance_idx(o[train], random_state=random_state)]
                if shuffle: train = random_shuffle(train, random_state)
                train_ = L(L([train]) * n_splits) if n_splits > 1 else train
                valid_ = L(L([train]) * n_splits) if n_splits > 1 else train
                test_ = L(L([test]) * n_splits) if n_splits > 1 else test
                if n_splits > 1: 
                    return [split for split in itemify(train_, valid_, test_)]
                else: 
                    return train_, valid_, test_
            elif n_splits > 1: 
                if stratify_: 
                    splits = StratifiedKFold(n_splits=n_splits, shuffle=shuffle, random_state=random_state).split(np.arange(len(train_valid)), o[train_valid])
                else:
                    splits = KFold(n_splits=n_splits, shuffle=shuffle, random_state=random_state).split(np.arange(len(train_valid)))
                train_, valid_ = L([]), L([])
                for train, valid in splits:
                    train, valid = toL(train), toL(valid)
                    if balance: train = train[balance_idx(o[train], random_state=random_state)]
                    if shuffle: 
                        train = random_shuffle(train, random_state)
                        valid = random_shuffle(valid, random_state)
                    train_.append(L(L(train_valid)[train]))
                    valid_.append(L(L(train_valid)[valid]))
                test_ = L(L([test]) * n_splits)
                return [split for split in itemify(train_, valid_, test_)]
            else:
                train, valid = train_test_split(range(len(train_valid)), test_size=vs, random_state=random_state, 
                                                stratify=o[train_valid] if stratify_ else None, shuffle=shuffle, **kwargs)
                train, valid = toL(train), toL(valid)
                if balance: train = train[balance_idx(o[train], random_state=random_state)]
                if shuffle: 
                    train = random_shuffle(train, random_state)
                    valid = random_shuffle(valid, random_state)
                return (L(L(train_valid)[train]), L(L(train_valid)[valid]),  test)
        else: 
            if vs == 0:
                train, _ = RandomSplitter(0, seed=random_state)(o)
                train = toL(train)
                if balance: train = train[balance_idx(o[train], random_state=random_state)]
                if shuffle: train = random_shuffle(train, random_state)
                train_ = L(L([train]) * n_splits) if n_splits > 1 else train
                valid_ = L(L([train]) * n_splits) if n_splits > 1 else train
                if n_splits > 1: 
                    return [split for split in itemify(train_, valid_)]
                else: 
                    return (train_, valid_)
            elif n_splits > 1: 
                if stratify_: splits = StratifiedKFold(n_splits=n_splits, shuffle=shuffle, random_state=random_state).split(np.arange(len(o)), o)
                else: splits = KFold(n_splits=n_splits, shuffle=shuffle, random_state=random_state).split(np.arange(len(o)))
                train_, valid_ = L([]), L([])
                for train, valid in splits:
                    train, valid = toL(train), toL(valid)
                    if balance: train = train[balance_idx(o[train], random_state=random_state)]
                    if shuffle: 
                        train = random_shuffle(train, random_state)
                        valid = random_shuffle(valid, random_state)
                    if not isinstance(train, (list, L)):  train = train.tolist()
                    if not isinstance(valid, (list, L)):  valid = valid.tolist()
                    train_.append(L(train))
                    valid_.append(L(L(valid)))
                return [split for split in itemify(train_, valid_)]
            else:
                train, valid = train_test_split(range(len(o)), test_size=vs, random_state=random_state, stratify=o if stratify_ else None, 
                                                shuffle=shuffle, **kwargs)
                train, valid = toL(train), toL(valid)
                if balance: train = train[balance_idx(o[train], random_state=random_state)]
                return train, valid
    return _inner

# %% ../../nbs/010_data.validation.ipynb 13
def plot_splits(splits):
    _max = 0
    _splits = 0
    for i, split in enumerate(splits):
        if is_listy(split[0]):
            for j, s in enumerate(split):
                _max = max(_max, array(s).max())
                _splits += 1
        else: 
            _max = max(_max, array(split).max())
            _splits += 1
    _splits = [splits] if not is_listy(split[0]) else splits
    v = np.zeros((len(_splits), _max + 1))
    for i, split in enumerate(_splits):
        if is_listy(split[0]):
            for j, s in enumerate(split): 
                v[i, s] = 1 + j
        else: v[i, split] = 1 + i
    vals = np.unique(v)
    plt.figure(figsize=(16, len(_splits)/2))
    if len(vals) == 1:
        v = np.ones((len(_splits), _max + 1))
        plt.pcolormesh(v, color='blue')
        legend_elements = [Patch(facecolor='blue', label='Train')]
        plt.legend(handles=legend_elements, bbox_to_anchor=(1.05, 1), loc='upper left')
    else: 
        colors = L(['gainsboro', 'blue', 'limegreen', 'red'])[vals]
        cmap = LinearSegmentedColormap.from_list('', colors)
        plt.pcolormesh(v, cmap=cmap)
        legend_elements = L([
            Patch(facecolor='gainsboro', label='None'),
            Patch(facecolor='blue', label='Train'),
            Patch(facecolor='limegreen', label='Valid'),
            Patch(facecolor='red', label='Test')])[vals]
        plt.legend(handles=legend_elements, bbox_to_anchor=(1.05, 1), loc='upper left')
    plt.title('Split distribution')
    plt.yticks(ticks=np.arange(.5, len(_splits)+.5, 1.0), labels=np.arange(1, len(_splits)+1, 1.0).astype(int))
    plt.gca().invert_yaxis()
    plt.show()

# %% ../../nbs/010_data.validation.ipynb 14
def get_splits(o, n_splits:int=1, valid_size:float=0.2, test_size:float=0., train_only:bool=False, train_size:Union[None, float, int]=None, balance:bool=False,
               shuffle:bool=True, stratify:bool=True, check_splits:bool=True, random_state:Union[None, int]=None, show_plot:bool=True, verbose:bool=False):
    '''Arguments: 
        o            : object to which splits will be applied, usually target.
        n_splits     : number of folds. Must be an int >= 1.
        valid_size   : size of validation set. Only used if n_splits = 1. If n_splits > 1 valid_size = (1. - test_size) / n_splits. 
        test_size    : size of test set. Default = 0.
        train_only   : if True valid set == train set. This may be useful for debugging purposes.
        train_size   : size of the train set used. Default = None (the remainder after assigning both valid and test). 
                        Useful for to get learning curves with different train sizes or get a small batch to debug a neural net.
        balance      : whether to balance data so that train always contain the same number of items per class.
        shuffle      : whether to shuffle data before splitting into batches. Note that the samples within each split will be shuffle.
        stratify     : whether to create folds preserving the percentage of samples for each class.
        check_splits : whether to perform leakage and completion checks.
        random_state : when shuffle is True, random_state affects the ordering of the indices. Pass an int for reproducible output.
        show_plot    : plot the split distribution
    '''
    if n_splits == 1 and valid_size == 0. and  test_size == 0.: train_only = True
    if balance: stratify = True
    splits = TrainValidTestSplitter(n_splits, valid_size=valid_size, test_size=test_size, train_only=train_only, stratify=stratify, 
                                      balance=balance, shuffle=shuffle, random_state=random_state, verbose=verbose)(o)
    if check_splits:
        if train_only or (n_splits == 1 and valid_size == 0): print('valid == train')
        elif n_splits > 1: 
            for i in range(n_splits): 
                leakage_finder([*splits[i]], verbose=True)
                cum_len = 0
                for split in splits[i]: cum_len += len(split)
                if not balance: assert len(o) == cum_len, f'len(o)={len(o)} while cum_len={cum_len}'
        else: 
            leakage_finder([splits], verbose=True)
            cum_len = 0
            if not isinstance(splits[0], Integral):
                for split in splits: cum_len += len(split)
            else: cum_len += len(splits)
            if not balance: assert len(o) == cum_len, f'len(o)={len(o)} while cum_len={cum_len}'
    if train_size is not None and train_size != 1: # train_size=1 legacy
        if n_splits > 1:
            splits = list(splits)
            for i in range(n_splits): 
                splits[i] = list(splits[i])
                if isinstance(train_size, Integral):
                    n_train_samples = train_size  
                elif train_size > 0 and train_size < 1: 
                    n_train_samples = int(len(splits[i][0]) * train_size)
                splits[i][0] = L(np.random.choice(splits[i][0], n_train_samples, False).tolist())
                if train_only:
                    if valid_size != 0: splits[i][1] = splits[i][0]
                    if test_size != 0: splits[i][2] = splits[i][0]
                splits[i] = tuple(splits[i])
            splits = tuple(splits)
        else: 
            splits = list(splits)
            if isinstance(train_size, Integral):
                n_train_samples = train_size  
            elif train_size > 0 and train_size < 1: 
                n_train_samples = int(len(splits[0]) * train_size)
            splits[0] = L(np.random.choice(splits[0], n_train_samples, False).tolist())
            if train_only:
                if valid_size != 0: splits[1] = splits[0]
                if test_size != 0: splits[2] = splits[0]
            splits = tuple(splits)
    if show_plot: plot_splits(splits)
    return splits

# %% ../../nbs/010_data.validation.ipynb 17
def get_walk_forward_splits(
    o, # 3D object with shape [samples x features x steps] containing the time series we need to split
    n_splits=1, # # of splits
    train_size=None, # optional: training set size as an int or a float. None when using and anchored strategy.
    valid_size=0.2, # validation set size as an int or a float
    test_size=0., # test set size as an int or a float
    anchored = False, # starting point for train set remains the same for all splits
    gap = 0., # # of samples to exclude from the end of each train set before the validation set. Entered as an int or a float
    test_after_valid = True, # flag to indicate if validation and test will be samples randomly or sequentially
    random_state = None, # integer that can be used to generate reproducible results
    show_plot=True, # plots the splits created
):

    if anchored:
        train_size = None
    elif isinstance(train_size, float): 
        train_size = np.int32(np.floor(len(o) * train_size))
    if isinstance(valid_size, float): 
        valid_size = np.int32(np.floor(len(o) * valid_size))
    if isinstance(test_size, float): 
        test_size = np.int32(np.floor(len(o) * test_size))
    if isinstance(gap, float): 
        gap = np.int32(np.floor(len(o) * gap))

    if train_size is not None:
        assert train_size + (valid_size + test_size + gap) * n_splits <= len(o), "reduce train_size, valid_size, test_size, gap or n_splits"
    else:
        assert (valid_size + test_size + gap) * n_splits < len(o), "reduce valid_size, test_size, gap or n_splits"

    if not test_after_valid:
        assert valid_size == test_size

    train_idxs = []
    valid_idxs = []
    test_idxs = []

    end = 0
    all_idxs = np.arange(len(o))
    for n in range(n_splits):
        if valid_size > 0 and test_size > 0:
            if test_after_valid:
                test_idxs.append(L(all_idxs[-test_size:].tolist()))
                all_idxs = all_idxs[:-test_size]
                valid_idxs.append(L(all_idxs[-valid_size:].tolist()))
                all_idxs = all_idxs[:-valid_size]
                if gap > 0:
                    all_idxs = all_idxs[:-gap]
                if anchored:
                    train_idxs.append(L(all_idxs.tolist()))
                else:
                    train_idxs.append(L(all_idxs[-train_size:].tolist()))
            else:
                valid_test_idxs = all_idxs[-test_size - valid_size:]
                np.random.seed(random_state)
                valid_test_idxs = np.random.permutation(valid_test_idxs)
                valid_idxs.append(L(valid_test_idxs[:valid_size]))
                test_idxs.append(L(valid_test_idxs[valid_size:]))
                all_idxs = all_idxs[:-test_size - valid_size]
                if gap > 0:
                    all_idxs = all_idxs[:-gap]
                if anchored:
                    train_idxs.append(L(all_idxs.tolist()))
                else:
                    train_idxs.append(L(all_idxs[-train_size:].tolist()))
        elif valid_size > 0:
            valid_idxs.append(L(all_idxs[-valid_size:].tolist()))
            all_idxs = all_idxs[:-valid_size]
            test_idxs.append(L([]))
            if gap > 0:
                all_idxs = all_idxs[:-gap]
            if anchored:
                train_idxs.append(L(all_idxs.tolist()))
            else:
                train_idxs.append(L(all_idxs[-train_size:].tolist()))

    splits = []
    for n in range(n_splits):
        if valid_size > 0 and test_size > 0:
            splits.append((L(train_idxs[n]), L(valid_idxs[n]), L(test_idxs[n])))
        elif valid_size > 0:
            splits.append((L(train_idxs[n]), L(valid_idxs[n])))
        else:
            splits.append((L(train_idxs[n]),))
    splits = tuple(splits)[::-1]
    if show_plot:
        plot_splits(splits)
    return splits

# %% ../../nbs/010_data.validation.ipynb 19
def TSSplitter(valid_size:Union[int, float]=0.2, test_size:Union[int, float]=0., show_plot:bool=True):
    "Create function that splits `items` between train/val with `valid_size` without shuffling data."
    def _inner(o):
        valid_cut = valid_size if isinstance(valid_size, Integral) else int(round(valid_size * len(o)))
        if test_size: 
            test_cut = test_size if isinstance(test_size, Integral) else int(round(test_size * len(o)))
        idx = np.arange(len(o))
        if test_size: 
            splits = L(idx[:-valid_cut - test_cut].tolist()), L(idx[-valid_cut - test_cut: - test_cut].tolist()), L(idx[-test_cut:].tolist())
        else: 
            splits = L(idx[:-valid_cut].tolist()), L(idx[-valid_cut:].tolist())
        if show_plot: 
            if len(o) > 1_000_000:
                warnings.warn('the splits are too large to be plotted')
            else: 
                plot_splits(splits)
        return splits
    return _inner

TimeSplitter = TSSplitter

# %% ../../nbs/010_data.validation.ipynb 36
def get_predefined_splits(*xs):
    '''xs is a list with X_train, X_valid, ...'''
    splits_ = []
    start = 0
    for x in xs: 
        splits_.append(L(list(np.arange(start, start + len(x)))))
        start += len(x)
    return tuple(splits_)

def combine_split_data(xs, ys=None):
    '''xs is a list with X_train, X_valid, .... ys is None or a list with y_train, y_valid, .... '''
    xs = [to3d(x) for x in xs]
    splits = get_predefined_splits(*xs)
    if ys is None: return concat(*xs), None, splits
    else: return concat(*xs), concat(*ys), splits

# %% ../../nbs/010_data.validation.ipynb 37
def get_splits_len(splits):
    _len = []
    for split in splits: 
        if isinstance(split[0], (list, L, tuple)):  _len.append([len(s) for s in split])
        else: _len.append(len(split))
    return _len



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/102_models.InceptionTime.ipynb.

# %% auto 0
__all__ = ['InceptionModule', 'InceptionBlock', 'InceptionTime']

# %% ../../nbs/102_models.InceptionTime.ipynb 3
from ..imports import *
from .layers import *

# %% ../../nbs/102_models.InceptionTime.ipynb 4
# This is an unofficial PyTorch implementation by Ignacio Oguiza - timeseriesAI@gmail.com based on:

# Fawaz, H. I., Lucas, B., Forestier, G., Pelletier, C., Schmidt, D. F., Weber, J., ... & Petitjean, F. (2019). 
# InceptionTime: Finding AlexNet for Time Series Classification. arXiv preprint arXiv:1909.04939.
# Official InceptionTime tensorflow implementation: https://github.com/hfawaz/InceptionTime

class InceptionModule(Module):
    def __init__(self, ni, nf, ks=40, bottleneck=True):
        ks = [ks // (2**i) for i in range(3)]
        ks = [k if k % 2 != 0 else k - 1 for k in ks]  # ensure odd ks
        bottleneck = bottleneck if ni > 1 else False
        self.bottleneck = Conv1d(ni, nf, 1, bias=False) if bottleneck else noop
        self.convs = nn.ModuleList([Conv1d(nf if bottleneck else ni, nf, k, bias=False) for k in ks])
        self.maxconvpool = nn.Sequential(*[nn.MaxPool1d(3, stride=1, padding=1), Conv1d(ni, nf, 1, bias=False)])
        self.concat = Concat()
        self.bn = BN1d(nf * 4)
        self.act = nn.ReLU()

    def forward(self, x):
        input_tensor = x
        x = self.bottleneck(input_tensor)
        x = self.concat([l(x) for l in self.convs] + [self.maxconvpool(input_tensor)])
        return self.act(self.bn(x))


@delegates(InceptionModule.__init__)
class InceptionBlock(Module):
    def __init__(self, ni, nf=32, residual=True, depth=6, **kwargs):
        self.residual, self.depth = residual, depth
        self.inception, self.shortcut = nn.ModuleList(), nn.ModuleList()
        for d in range(depth):
            self.inception.append(InceptionModule(ni if d == 0 else nf * 4, nf, **kwargs))
            if self.residual and d % 3 == 2: 
                n_in, n_out = ni if d == 2 else nf * 4, nf * 4
                self.shortcut.append(BN1d(n_in) if n_in == n_out else ConvBlock(n_in, n_out, 1, act=None))
        self.add = Add()
        self.act = nn.ReLU()
        
    def forward(self, x):
        res = x
        for d, l in enumerate(range(self.depth)):
            x = self.inception[d](x)
            if self.residual and d % 3 == 2: res = x = self.act(self.add(x, self.shortcut[d//3](res)))
        return x

    
@delegates(InceptionModule.__init__)
class InceptionTime(Module):
    def __init__(self, c_in, c_out, seq_len=None, nf=32, nb_filters=None, **kwargs):
        nf = ifnone(nf, nb_filters) # for compatibility
        self.inceptionblock = InceptionBlock(c_in, nf, **kwargs)
        self.gap = GAP1d(1)
        self.fc = nn.Linear(nf * 4, c_out)

    def forward(self, x):
        x = self.inceptionblock(x)
        x = self.gap(x)
        x = self.fc(x)
        return x



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/121b_models.GatedTabTransformer.ipynb.

# %% auto 0
__all__ = ['GatedTabTransformer']

# %% ../../nbs/121b_models.GatedTabTransformer.ipynb 4
import torch
import torch.nn as nn

from .TabTransformer import TabTransformer
from .gMLP import gMLP

# %% ../../nbs/121b_models.GatedTabTransformer.ipynb 5
class _TabMLP(nn.Module):
    def __init__(self, classes, cont_names, c_out, d_model, mlp_d_model, mlp_d_ffn, mlp_layers):
        super().__init__()
        seq_len = d_model * len(classes) + len(cont_names)
        self.mlp = gMLP(1, c_out, seq_len, d_model=mlp_d_model, d_ffn=mlp_d_ffn, depth=mlp_layers)
    
    def forward(self, x):
        x = x.unsqueeze(1)
        return self.mlp(x)


class GatedTabTransformer(TabTransformer):
    def __init__(self, classes, cont_names, c_out, column_embed=True, add_shared_embed=False, shared_embed_div=8, embed_dropout=0.1, drop_whole_embed=False, 
                 d_model=32, n_layers=6, n_heads=8, d_k=None, d_v=None, d_ff=None, res_attention=True, attention_act='gelu', res_dropout=0.1, norm_cont=True,
                 mlp_d_model=32, mlp_d_ffn=64, mlp_layers=4):

        super().__init__(classes, cont_names, c_out, column_embed=column_embed, add_shared_embed=add_shared_embed, shared_embed_div=shared_embed_div,
                         embed_dropout=embed_dropout, drop_whole_embed=drop_whole_embed, d_model=d_model, n_layers=n_layers, n_heads=n_heads, d_k=d_k,
                         d_v=d_v, d_ff=d_ff, res_attention=res_attention, attention_act=attention_act, res_dropout=res_dropout, norm_cont=norm_cont)

        self.mlp = _TabMLP(classes, cont_names, c_out, d_model, mlp_d_model, mlp_d_ffn, mlp_layers)
